import torch

PITCH_MAX = 128
VELO = 128
LENGTH = 999
LENGTH_HALF = 999
BEGIN = 999
BEGIN_HALF = 999
ROOT = 99
START_SEQ_TOKEN = "<S_SEQ>"
END_SEQ_TOKEN = "<E_SEQ>"
PADDING_TOKEN = "<PAD>"

MODEL_NAME = "MORTM"

# 前回のトークンのID ＋　前回のトークンの使用個数
PADDING_BEGIN_ID = 0
SPECIAL_BEGIN_ID = PADDING_BEGIN_ID + 1
PITCH_BEGIN_ID = SPECIAL_BEGIN_ID + 2
VELOCITY_BEGIN_ID = PITCH_BEGIN_ID + 128
DURATION_BEGIN_ID = VELOCITY_BEGIN_ID + 128
START_BEGIN_ID = DURATION_BEGIN_ID + 100
SHIFT_BEGIN_ID = START_BEGIN_ID + 32


PITCH_GROUP = range(PITCH_BEGIN_ID, PITCH_BEGIN_ID + 128 + 1)
VELOCITY_GROUP = range(VELOCITY_BEGIN_ID, VELOCITY_BEGIN_ID + 128 + 1)
DURATION_GROUP = range(DURATION_BEGIN_ID, DURATION_BEGIN_ID + 100 + 1)
START_GROUP = range(START_BEGIN_ID, START_BEGIN_ID + 32 + 1)
SHIFT_GROUP = range(SHIFT_BEGIN_ID, SHIFT_BEGIN_ID + 4 + 1)




from AGSM.convert import ConvTempo
from pretty_midi.pretty_midi import PrettyMIDI, Instrument, Note
import mido
import numpy as np
from numpy import ndarray
from typing import List, Callable

from . import constants
from . import aya_node
from .tokenizer import Tokenizer


def _get_token_converter(tempo:int, tokenizer: Tokenizer) -> List[Callable[[Note, Note], int]]:
    register: List[Callable[[Note, Note], int]] = list()

    register.append(aya_node.get_token(tempo, tokenizer, aya_node.SHIFT_TYPE))
    register.append(aya_node.get_token(tempo, tokenizer, aya_node.START_TYPE))
    register.append(aya_node.get_token(tempo, tokenizer, aya_node.PITCH_TYPE))
    register.append(aya_node.get_token(tempo, tokenizer, aya_node.VELOCITY_TYPE))
    register.append(aya_node.get_token(tempo, tokenizer, aya_node.DURATION_TYPE))

    return register


class MidiToAyaNode:

    def __init__(self, tokenizer: Tokenizer, directory: str, file_name: str, program_list, midi_data=None,
                 token_converter: List[Callable[[Note, Note], int]]=None):
        self.program_list = program_list
        self.tokenizer = tokenizer
        self.directory = directory
        self.file_name = file_name
        self.tempo = 120
        self.is_error = False
        self.aya_node = [0]
        if token_converter is not None:
            self.token_converter: List[Callable[[Note, Note], int]] = token_converter
        else:
            self.token_converter: List[Callable[[Note, Note], int]] = _get_token_converter(self.tempo, tokenizer)

        if midi_data is not None:
            self.midi_data: PrettyMIDI = midi_data
        else:
            try:
                self.midi_data: PrettyMIDI = PrettyMIDI(f"{directory}/{file_name}")
            except Exception:
                self.is_error = True
                print("ロードできませんでした。")
        pass

    def convert(self):
        """
        以下のステップでMidiが変換される
        1. テンポ固定
            120テンポに変換される。
        2. 指定した楽器があるかを検索する
            インスタンス化した際に、指定したprogram_listに乗っ取り、該当する楽器があるかを検索する
        3. ノートを変換する。
            self.ct_aya_nodeにインストを渡すことで、MORTMに適した形に変換する。

        また、変換したインストゥルメントは以下のような配列構造になる。

        1_music = [
                    [1 inst's 60s clip],
                        ...
                    [1 inst's 60s clip],
                    [2 inst's 60s clip],
                        ...
                    [n inst's 60s clip]
                    ]
        :return:なし。
        """
        if not self.is_error:
            self.ct_tempo()
            program_count = 0
            for inst in self.midi_data.instruments:
                inst: Instrument = inst
                if not inst.is_drum and inst.program in self.program_list:
                    print(f"Instrument Number:{inst.program}")
                    aya_node_inst = self.ct_aya_node(inst)

                    self.aya_node = self.aya_node + aya_node_inst
                    #print(self.aya_node)
                    program_count += 1

            if program_count == 0:
                self.is_error = True
                print(f"{self.directory}/{self.file_name}に、欲しい楽器がありませんでした。")
        else:
            print(f"{self.directory}/{self.file_name}を変換できません。")


    def ct_aya_node(self, inst: Instrument) -> list:

        """
        Instrumentsから1音ずつ取り出し、Tokenizerで変換する。
        clip = [<START>, S, P, V, D, H, S, P, V, D, H ...<END>]
        さらに60秒ごとにスプリットし、以下のような配列構造を作る。
        aya_node_inst = [clip_1, clip_2 ... clip_n]

        よって、二次元配列のndarrayを返す。

        :param inst: インストゥルメント
        :return: 60秒にクリッピングされた旋律の配列(2次元)
        """

        clip = np.array([], dtype=int)
        aya_node_inst = []
        back_note = None

        clip_time: float = 0
        split_count: int = 1

        for note in inst.notes:
            note: Note = note

            if back_note is None:
                clip = np.append(clip, self.tokenizer.get(-1, constants.START_SEQ_TOKEN))

            for conv in self.token_converter:
                conv: Callable[[Note, Note], int] = conv

                token = conv(back_note, note)
                if token != -1:
                    clip = np.append(clip, token)

            back_note = note

            clip_time = note.end

            if clip_time >= 60 * split_count:
                if len(clip) >= 10:
                    clip = np.append(clip, self.tokenizer.get(-1, constants.END_SEQ_TOKEN))

                    aya_node_inst.append(clip)

                clip = np.array([], dtype=int)
                back_note = None

                clip_time = 0.0
                split_count += 1

        if len(clip) >= 10:
            clip = np.append(clip, self.tokenizer.get(-1, constants.END_SEQ_TOKEN))
            aya_node_inst.append(clip)

        return aya_node_inst


    def save(self, save_directory: str) -> bool:
        if not self.is_error:
            #print(f"Result shape is:{self.aya_node.shape}")

            array_dict = {f'array{i}': arr for i, arr in enumerate(self.aya_node)}
            np.savez(save_directory + "/" + self.file_name, **array_dict)
            print("処理が正常に終了しました。")
            return True
        else:
            print(f"{constants.MODEL_NAME}が望むデータ形式ではないため、保存ができませんでした。")
            return False

    def ct_tempo(self):
        try:
            ct = ConvTempo(directory=self.directory + "/" + self.file_name, midi_data=self.midi_data, change_tempo=120)
            ct.convert()
            self.midi_data = ct.midi_data
        except OSError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True
        except IndexError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True
        except ValueError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True

        except mido.midifiles.meta.KeySignatureError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True

        except EOFError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True

        except KeyError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True
        except ZeroDivisionError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True


class AyaNodeToMidi:
    def __init__(self, directory:str):
        self.directory = directory
        npz = np.load(directory)
        self.npz_dict = npz
        self.midi_data = None
        print(self.npz_dict)

    def convert(self):
        pass


import numpy as np
from numpy import ndarray
from pretty_midi.pretty_midi import Note
from typing import Callable

from .tokenizer import Tokenizer
from .tokenizer import PITCH_TYPE,SHIFT_TYPE,START_TYPE,VELOCITY_TYPE,DURATION_TYPE


def ct_time_to_beat(time: float, tempo: int) -> int:
    b4 = 60 / tempo
    b8 = b4 / 2
    b16 = b8 / 2
    b32 = b16 / 2

    beat, sub = calc_time_to_beat(time, b32)

    return beat


def calc_time_to_beat(time, beat_time) -> (int, int):
    main_beat: int = time // beat_time
    sub_time: int = time % beat_time
    return main_beat, sub_time


def get_token(tempo: int, tokenizer: Tokenizer, token_type: str) -> Callable[[Note, Note], int]:
    def get_pitch(back_notes: Note, note: Note) -> int:
        p: int = note.pitch
        return tokenizer.get(p, PITCH_TYPE)



    def get_velocity(back_notes: Note, note: Note) -> int:
        v: int = note.velocity
        return tokenizer.get(v, VELOCITY_TYPE)


    def get_duration(back_note: Note, note: Note) -> int:
        start = ct_time_to_beat(note.start, tempo)
        end = ct_time_to_beat(note.end, tempo)
        d = int(max(abs(end - start), 1))

        if 100 < d:
            d = 100

        return tokenizer.get(d, DURATION_TYPE)


    def get_start(back:Note, note: Note) -> int:
        s = ct_time_to_beat(note.start, tempo)

        return tokenizer.get(int(s % 32), START_TYPE)


    def get_shift(back: Note, note: Note) -> int:
        if back is None:
            return -1
        else:
            back_start = ct_time_to_beat(back.start, tempo)
            note_start = ct_time_to_beat(note.start, tempo)

            shift = int(abs((back_start // 32) - (note_start // 32)))

            if shift > 3:
                shift = 3

            return tokenizer.get(shift, SHIFT_TYPE)


    if token_type is PITCH_TYPE:
        return get_pitch
    elif token_type is VELOCITY_TYPE:
        return get_velocity
    elif token_type is DURATION_TYPE:
        return get_duration
    elif token_type is START_TYPE:
        return get_start
    elif token_type is SHIFT_TYPE:
        return get_shift
